import collections
import json
import queue
import time

import networkx

from appium import webdriver
from appium.webdriver.common.appiumby import AppiumBy
from selenium.common.exceptions import InvalidElementStateException

from ..exploration_strategy import ExplorationStrategy


class iOSStrategy(ExplorationStrategy):
    class Meta:
        platforms: list[str] = ['iPhone OS']

    class XCUIElement:
        def __init__(self, properties_dict):
            for key in properties_dict:
                if key != 'children':
                    setattr(self, key, properties_dict[key])

        def __eq__(self, other):
            comparison_values = ['type', 'rect', 'name', 'label']
            for attribute_name in comparison_values:
                if getattr(self, attribute_name, None) != getattr(other, attribute_name, None):
                    return False
            return True

        def __lt__(self, other):
            return self.rect('y'), self.rect('x') < other.rect('y'), other.rect('x')

        def __hash__(self):
            return hash((frozenset(self.rect),
                         self.label if self.label else "",
                         self.name if self.name else "",
                         self.value if self.value else ""
                         ))

        def get_wd_element(self, wd: webdriver):
            return wd.find_element(
                by=AppiumBy.IOS_PREDICATE,
                value=
                f"type == 'XCUIElementType{self.type}' AND "
                f"rect.x == {self.rect['x']} AND "
                f"rect.y == {self.rect['y']}"
            )

        INTERACTIONS = {
            'tap': ['Button']
        }
        ELEMENT_INTERACTIONS = {
            'Button': ['tap']
        }

    def interact(self):
        graph = self.parse_json_source()

    def parse_json_source(self) -> networkx.DiGraph:
        IGNORE_TYPES = [
            'Keyboard',
            'StatusBar'
        ]

        graph = networkx.DiGraph()

        try:
            json_source = self.wd.execute_script('mobile:source', {'format': 'json'})
        except InvalidElementStateException as ex:
            self.logger.error(f'Failed to get UI source representation: {ex.msg}')
            return graph
        children = json_source.pop('children', [])
        root_el = self.XCUIElement(json_source)
        graph.add_node(root_el)

        q: queue.Queue = queue.Queue()
        for child in children:
            q.put((child, root_el))
        while not q.empty():
            (el, parent) = q.get()
            if el.get('type', '') in IGNORE_TYPES:
                continue
            children = el.pop('children', [])
            el = self.XCUIElement(el)
            graph.add_node(el)
            graph.add_edge(parent, el)
            for child in children:
                q.put((child, el))

        return graph



import subprocess

from .cmdline_util import CmdlineUtil, CmdlineUtilUnavailableException
from ...schemas import DeviceConnected


class Ideviceinfo(CmdlineUtil):
    name = "ideviceinfo"

    @staticmethod
    def run(device: DeviceConnected, keys: list[str]) -> dict[str, str]:
        result = {}
        for key in keys:
            try:
                ideviceinfo: subprocess.CompletedProcess = subprocess.run(
                    args=[Ideviceinfo.name, "-u", device.udid, "-k", key],
                    capture_output=True,
                    check=True
                )
                result[key] = ideviceinfo.stdout.rstrip().decode("UTF-8")
            except subprocess.CalledProcessError:
                pass
            except FileNotFoundError:
                raise CmdlineUtilUnavailableException
        return result




import logging
import os
import re
import zipfile
from os import path
import plistlib

from .. import schemas


def scan_directory(base_path: path):
    ipas: list[schemas.IPA] = []
    with os.scandir(base_path) as files:
        for f in filter(lambda f: os.path.basename(f).endswith('.ipa'), files):
            try:
                plist = None

                if zipfile.is_zipfile(f):
                    with zipfile.ZipFile(f) as ipa_zip:
                        info_plist_pattern = re.compile(r"Payload/.*\.app/Info\.plist")
                        info_plist_pathstr: str = next(
                            filter(lambda x: info_plist_pattern.match(x), ipa_zip.namelist()))
                        with ipa_zip.open(info_plist_pathstr) as info_plist:
                            plist = plistlib.load(info_plist)
                else:
                    with open(os.path.join(f, 'Info.plist'), 'rb') as fp:
                        plist = plistlib.load(fp)

                ipa: schemas.IPA = schemas.IPA(path=f,
                                               bundle_id=plist['CFBundleIdentifier'],
                                               minimum_os_version=plist['MinimumOSVersion'],
                                               supported_platforms=plist['CFBundleSupportedPlatforms'],
                                               bundle_name=plist['CFBundleName'],
                                               bundle_version=plist['CFBundleVersion']
                                               )
                ipas.append(ipa)
            except FileNotFoundError:
                continue
            except NotADirectoryError:
                continue
            except KeyError as e:
                logging.log(logging.WARNING, f"{f}/Info.plist is missing required key: {e}")
                continue
    return ipas



from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.encoders import jsonable_encoder
from fastapi.responses import HTMLResponse
from sqlalchemy.orm import Session

from ..dependencies import get_db, templates
from ... import crud, schemas


router = APIRouter()


@router.get("/appium-servers", response_class=HTMLResponse)
async def appium_servers(request: Request, db: Session = Depends(get_db)):
    servers = get_appium_servers(db=db)
    return templates.TemplateResponse("appium-servers.html", {
        "request": request,
        "servers": jsonable_encoder(servers)
    })


@router.get("/api/appium_servers", response_model=list[schemas.AppiumServer])
def get_appium_servers(skip: int = 0, limit: int = 100, db: Session = Depends(get_db)):
    servers = crud.get_appium_servers(db, skip=skip, limit=limit)
    return servers


@router.get("/api/appium_servers/{id}", response_model=schemas.AppiumServer)
def get_appium_server_by_id(id: int, db: Session = Depends(get_db)):
    server = crud.get_appium_server_by_id(db, id=id)
    if server is None:
        raise HTTPException(status_code=404, detail="Server not found")
    return server


@router.post("/api/appium_servers", response_model=schemas.AppiumServer)
def add_appium_server(appium_server_create: schemas.AppiumServerCreate, db: Session = Depends(get_db)):
    appium_server = crud.add_appium_server(db, appium_server=appium_server_create)
    return appium_server


@router.delete("/api/appium_servers/{id}", response_model=schemas.AppiumServer)
def delete_appium_by_server_id(id: int, db: Session = Depends(get_db)):
    return crud.delete_appium_server(db, appium_server_id=id)




import os

from fastapi.templating import Jinja2Templates

from ..database import SessionLocal


def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()


templates = Jinja2Templates(directory=os.path.join(os.path.abspath(os.path.dirname(__file__)), os.path.join("templates")))



import re
import subprocess

from .cmdline_util import CmdlineUtil, CmdlineUtilUnavailableException


class RvictlError(Exception):
    pass


class Rvictl(CmdlineUtil):
    name: str = "rvictl"

    @staticmethod
    def start_device(udid: str) -> str:
        success_re = re.compile(br"Starting device \w+ \[SUCCEEDED] with interface (?P<udid>\w+)")
        failure_re = re.compile(f"Starting device {udid} \[FAILED\]")

        try:
            p: subprocess.CompletedProcess = subprocess.run(
                args=[Rvictl.name, "-s", udid],
                capture_output=True,
                check=True
            )
            match = success_re.search(p.stdout)
            if not match:
                raise RvictlError
            return match.group("udid")
        except subprocess.CalledProcessError as e:
            raise RvictlError
        except FileNotFoundError as e:
            raise CmdlineUtilUnavailableException

    @staticmethod
    def stop_device(udid: str) -> None:
        try:
            p: subprocess.CompletedProcess = subprocess.run(
                args=["rvictl", "-x", udid],
                capture_output=True,
                check=True
            )
        except subprocess.CalledProcessError as e:
            raise RvictlError
        except FileNotFoundError as e:
            raise CmdlineUtilUnavailableException

    @staticmethod
    def list_devices():

        try:
            p: subprocess.CompletedProcess = subprocess.run(
                args=["rvictl", "-l"],
                capture_output=True,
                check=True
            )
        except subprocess.CalledProcessError as e:
            raise RvictlError
        except FileNotFoundError as e:
            raise CmdlineUtilUnavailableException



import logging
import threading
import time


from .exploration_session_options import ExplorationSessionOptions
from .exploration_session import ExplorationSession, ServerUnavailableError, ExplorationFailedError, ExplorationStartFailedError
from .. import models
from ..common import grievios_base_path
from ..database import SessionLocal


class ExplorationWorker:
    INTERVAL: int = 10

    def __init__(self, app, appium_server_id: int):
        self.app = app
        self.appium_server_id: int = appium_server_id
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def run(self):
        while not self._stop_event.is_set():
            db = SessionLocal()
            server = db.query(models.AppiumServer).filter(models.AppiumServer.id == self.appium_server_id).first()
            if not server:
                self.stop()

            exploration = db.query(models.Exploration).filter(
                models.Exploration.appium_server_id == self.appium_server_id,
                models.Exploration.status == "Created")\
                .first()
            if not exploration:
                logging.log(logging.DEBUG, f"No exploration for server {server.id} waiting, sleeping for {self.INTERVAL}s")
                db.close()
                time.sleep(self.INTERVAL)
                continue

            session_options: ExplorationSessionOptions = ExplorationSessionOptions(
                command_executor=f"{server.url}:{server.port}",  # TODO: server.basePath
                device_udid=exploration.device_udid,
                bundle_id=exploration.bundle_id,
                wda_bundle_id=server.wdaBundleId,
                strategy=exploration.strategy,
                analyzers=exploration.analyzers,
                log_directory=grievios_base_path() / "logs" / f"exploration-{str(exploration.id)}",
                bundle_installed=exploration.bundle_on_device
            )

            try:
                with ExplorationSession(self.app, session_options) as sess:
                    sess.start()
            except ServerUnavailableError:
                logging.log(logging.ERROR, "Appium server not reachable")
                time.sleep(self.INTERVAL)
                continue
            except ExplorationFailedError as e:
                logging.log(logging.ERROR, f'Exploration failed: {e}')
                exploration.status = 'Failed'
                db.commit()
                continue
            except ExplorationStartFailedError as e:
                logging.log(logging.ERROR, f'Exploration failed to start: {e.cause}')
                exploration.status = 'Failed'
                db.commit()
                continue

            except Exception as e:
                logging.log(logging.ERROR, f'Exploration failed unexpectedly: {e}')
                exploration.status = 'Failed'
                db.commit()
                continue

            exploration.status = 'Finished'
            db.commit()

            db.close()



import logging

from fastapi import APIRouter, Body, Depends, HTTPException, Request
from fastapi.encoders import jsonable_encoder
from fastapi.responses import HTMLResponse
from sqlalchemy.orm import Session

from ..dependencies import get_db, templates
from ... import analyzers, strategies
from ... import crud, schemas, common
from ...utils import ipa

router = APIRouter()


@router.get("/explorations", response_class=HTMLResponse)
async def explorations(request: Request, db: Session = Depends(get_db)):
    db_explorations = crud.get_explorations(db=db)
    db_devices = crud.get_devices(db=db)
    db_servers = crud.get_appium_servers(db=db)
    local_ipas = ipa.scan_directory(common.grievios_base_path() / 'IPAs')
    return templates.TemplateResponse("explorations.html", {
        "request": request,
        "explorations": jsonable_encoder(db_explorations),
        "devices": jsonable_encoder(db_devices),
        "servers": jsonable_encoder(db_servers),
        "strategies": strategies.ios.availableStrategies(),
        "analyzers": analyzers.ios.availableAnalyzers(),
        "localIpas": jsonable_encoder(local_ipas)
    })


@router.get("/api/explorations", response_model=list[schemas.Exploration])
def get_explorations(skip: int = 0, limit: int = 100, db: Session = Depends(get_db)):
    db_explorations = crud.get_explorations(db, skip=skip, limit=limit)
    return db_explorations


@router.post("/api/explorations", response_model=list[schemas.Exploration])
def add_explorations(explorationInput: schemas.ExplorationInput, db: Session = Depends(get_db)):
    try:
        device = crud.get_device(db, device_udid=explorationInput.device)
    except:
        raise HTTPException(status_code=404, detail=f"Device {explorationInput.device} not found")
    try:
        server = crud.get_appium_server_by_id(db, id=explorationInput.appiumServer)
    except:
        raise HTTPException(status_code=404, detail="Server not found")
    db_explorations: list[schemas.Exploration] = []
    for strategy in explorationInput.strategies:
        for localApp in explorationInput.localApps:
            try:
                newExploration = schemas.ExplorationCreate(
                    device_udid=device.udid,
                    appium_server_id=server.id,
                    status="Created",
                    bundle_id=localApp,
                    bundle_on_device=False,
                    strategy=strategy,
                    analyzers=explorationInput.analyzers
                )
                db_exploration = crud.add_exploration(db, newExploration)
                db_explorations.append(db_exploration)
            except Exception as e:
                logging.log(logging.ERROR, f"Error adding new exploration: {e}")
                continue

    return db_explorations


@router.put("/api/explorations/{exploration_id}/status/reset", response_model=schemas.Exploration)
def update_exploration(exploration_id: int, db: Session = Depends(get_db)):
    try:
        exploration: schemas.Exploration = crud.get_exploration_by_id(db, exploration_id)
    except:
        raise HTTPException(status_code=404, detail=f"Exploration {exploration_id} not found")
    exploration.status = 'Created'
    db.commit()
    return exploration


@router.delete("/api/explorations/{exploration_id}")
def delete_exploration_by_id(exploration_id: int, db: Session = Depends(get_db)):
    return crud.delete_exploration(db, exploration_id=exploration_id)



import time

from .ios_strategy import iOSStrategy


class IdleStrategy(iOSStrategy):

    def interact(self):
        self.logger.info(f"Idling for 1 second")
        time.sleep(1)
        self.logger.info(f"Done idling")




from dataclasses import dataclass
import re
import subprocess

from .cmdline_util import CmdlineUtil, CmdlineUtilUnavailableException
from ...schemas import DeviceConnected


@dataclass
class IDevice:
    udid: str
    name: str
    os: str
    os_version: str


class IdeviceId(CmdlineUtil):
    name = "idevice_id"

    @staticmethod
    def run() -> list[DeviceConnected]:
        devices: list[DeviceConnected] = []
        try:
            idevice_ids: subprocess.CompletedProcess = subprocess.run(
                args=[IdeviceId.name],
                capture_output=True,
                check=True
            )
        except subprocess.CalledProcessError:
            return []
        except FileNotFoundError:
            raise CmdlineUtilUnavailableException
        for line in idevice_ids.stdout.split(b'\n'):
            re_result = re.search(r'^([a-zA-Z0-9\-]+).+\([USB]+\)$', line.decode('utf-8'))
            if re_result:
                udid = re_result.groups()[0]
                devices.append(DeviceConnected(udid=udid))

        return devices



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from .common import grievios_base_path

DB_FILE = grievios_base_path() / 'grievios.db'
SQLALCHEMY_DATABASE_URL = f"sqlite:///{DB_FILE}"

engine = create_engine(
    SQLALCHEMY_DATABASE_URL, connect_args={"check_same_thread": False}
)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

Base = declarative_base()



import datetime
import logging
import os
from typing import Type

from appium import webdriver
from appium.options.ios import XCUITestOptions
from selenium.common.exceptions import WebDriverException
from urllib3.exceptions import MaxRetryError

from .exploration_session_options import ExplorationSessionOptions
from ..common import grievios_base_path
from .. import analyzers
from .. import strategies
from ..utils.cmdline_utils import ideviceinstaller, Ideviceinstaller
from ..utils.cmdline_utils.cmdline_util import CmdlineUtilUnavailableException


class ServerUnavailableError(Exception):
    pass


class ExplorationFailedError(Exception):
    pass


class ExplorationStartFailedError(Exception):
    cause: Exception

    def __init__(self, cause: Exception):
        self.cause = cause


class ExplorationSession:
    analyzer_instance: analyzers.ios.ios_analyzer.iOSAnalyzer = None
    analyzer_instances: list[analyzers.ios.ios_analyzer.iOSAnalyzer] = None

    def __init__(self, app, options: ExplorationSessionOptions):
        self.app = app
        self.options = options
        self.appium_wd: webdriver = None
        self.logger = logging.getLogger(os.path.basename(self.options.log_directory))

    def __enter__(self):
        try:
            os.makedirs(self.options.log_directory, exist_ok=True)
        except OSError as e:
            raise ExplorationStartFailedError(cause=e)
        self.logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s %(levelname)s in %(module)s - %(message)s')
        file_logger = logging.FileHandler(
            os.path.join(self.options.log_directory, datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S") + '.log'))
        file_logger.setFormatter(formatter)
        self.logger.addHandler(file_logger)

        if not self.options.bundle_installed:
            try:
                self.logger.info(f'Installing bundle {self.options.bundle_id} IPA onto {self.options.device_udid}')
                Ideviceinstaller.install_ipa(self.options.device_udid,
                                             grievios_base_path() / 'IPAs' / f'{self.options.bundle_id}.ipa')
            except ideviceinstaller.InstallationError as e:
                self.logger.error(f'Failed to install {self.options.bundle_id} IPA')
                raise ExplorationStartFailedError(cause=e)
            except CmdlineUtilUnavailableException as e:
                self.logger.error(f"ideviceinstaller is required for app installation")
                raise ExplorationStartFailedError(cause=e)

        analyzerClasses: filter[Type[analyzers.ios.ios_analyzer.iOSAnalyzer]] = filter(
            lambda c: c.__name__ in self.options.analyzers,
            analyzers.ios.ios_analyzer.iOSAnalyzer.__subclasses__()
        )
        self.analyzer_instances = [analyzerClass(exploration_session_options=self.options) for analyzerClass in
                                   analyzerClasses]

        options = XCUITestOptions()
        options.udid = self.options.device_udid
        options.bundle_id = self.options.bundle_id
        options.updated_wda_bundle_id = self.options.wda_bundle_id
        try:
            self.appium_wd = webdriver.Remote(self.options.command_executor, options=options)
        except MaxRetryError:
            raise ServerUnavailableError
        except WebDriverException as e:
            raise ExplorationStartFailedError(e)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.appium_wd:
            self.logger.info(f'Quitting app {self.options.bundle_id}')
            self.appium_wd.execute_script('mobile: terminateApp', {'bundleId': self.options.bundle_id})
            self.logger.info(f'Quitting Webdriver connection')
            self.appium_wd.quit()

            if not self.options.bundle_installed:
                self.logger.info(f'Uninstalling bundle {self.options.bundle_id} IPA')
                Ideviceinstaller.uninstall_ipa(self.options.device_udid, self.options.bundle_id)

            self.analyzer_instances = []

            for handler in self.logger.handlers:
                handler.close()

    def start(self):
        self.logger.info(f'Starting exploration session')

        strategyClass: Type[strategies.ios.ios_strategy.iOSStrategy] = next(filter(
            lambda c: c.__name__ == self.options.strategy,
            strategies.ios.ios_strategy.iOSStrategy.__subclasses__()),
            strategies.ios.ios_strategy.iOSStrategy)
        strategy_instance = strategyClass(wd=self.appium_wd, log_directory=self.options.log_directory,
                                          timeout=self.options.timeout)
        strategy_instance.run()



import pathlib

from pydantic import BaseModel, validator

from . import models


class DeviceBase(BaseModel):
    udid: str
    name: str = ""
    os: str = ""
    os_version: str = ""


class DeviceCreate(DeviceBase):
    pass


class Device(DeviceBase):
    name: str
    os: str
    os_version: str

    class Config:
        orm_mode = True
        orm_model = models.Device


class DeviceConnected(DeviceBase):
    pass


class AppiumServerBase(BaseModel):
    url: str
    port: int
    wdaBundleId: str


class AppiumServerCreate(AppiumServerBase):
    pass


class AppiumServer(AppiumServerBase):
    id: int

    class Config:
        orm_mode = True
        orm_model = models.AppiumServer


class ExplorationBase(BaseModel):
    #device: Device
    #appium_server: AppiumServer
    status: str
    bundle_id: str
    bundle_on_device: bool
    strategy: str
    analyzers: list[str] = []


class ExplorationCreate(ExplorationBase):
    device_udid: str
    appium_server_id: int
    pass


class Exploration(ExplorationBase):
    id: int
    device: Device
    appium_server: AppiumServer

    class Config:
        orm_mode = True
        orm_model = models.Exploration


class ExplorationInput(BaseModel):
    appiumServer: int
    device: str
    strategies: list[str]
    analyzers: list[str]
    installedApps: list[str]
    localApps: list[str]


class IPA(BaseModel):
    bundle_id: str
    minimum_os_version: str = None
    supported_platforms: list[str] = None
    bundle_name: str = None
    display_name: str = None
    bundle_version: str = None
    app_icon_path: pathlib.Path = None
    path: pathlib.Path = None



import os
import subprocess

from .cmdline_util import CmdlineUtil, CmdlineUtilUnavailableException


class InstallationError(Exception):
    pass


class UninstallationError(Exception):
    pass


class Ideviceinstaller(CmdlineUtil):
    name = "ideviceinstaller"

    @staticmethod
    def install_ipa(udid: str, path: os.path):
        try:
            p: subprocess.CompletedProcess = subprocess.run(
                args=[Ideviceinstaller.name, '-u', udid, '-i', path],
                capture_output=True,
                check=True
            )
        except subprocess.CalledProcessError as e:
            raise InstallationError
        except FileNotFoundError as e:
            raise CmdlineUtilUnavailableException

    @staticmethod
    def uninstall_ipa(udid: str, bundle_id: str):
        try:
            p: subprocess.CompletedProcess = subprocess.run(
                args=[Ideviceinstaller.name, '-u', udid, '-U', bundle_id],
                capture_output=True,
                check=True
            )
        except subprocess.CalledProcessError as e:
            raise UninstallationError
        except FileNotFoundError:
            raise UninstallationError




from dataclasses import dataclass
import os

@dataclass
class ExplorationSessionOptions:
    command_executor: str
    device_udid: str
    bundle_id: str
    wda_bundle_id: str
    strategy: str
    log_directory: os.path
    analyzers: list[str] = None
    benchmarks:str = None
    timeout: int = 55
    bundle_installed: bool = False



from sqlalchemy import Boolean, Column, Integer, Text, PickleType, ForeignKey
from sqlalchemy.orm import relationship, mapped_column

from .database import Base


class Device(Base):
    __tablename__ = "devices"

    udid = Column(Text, primary_key=True, nullable=False)
    name = Column(Text, nullable=False)
    os = Column(Text, nullable=False)
    os_version = Column(Text, nullable=False)


class AppiumServer(Base):
    __tablename__ = "appium_servers"

    id = Column(Integer, primary_key=True, autoincrement=True)
    url = Column(Text, nullable=False)
    port = Column(Integer, nullable=False)
    wdaBundleId = Column(Text)


class Exploration(Base):
    __tablename__ = "explorations"

    id = Column(Integer, primary_key=True, autoincrement=True)
    device_udid = mapped_column(ForeignKey("devices.udid"))
    appium_server_id = mapped_column(ForeignKey("appium_servers.id"))
    status = Column(Text, nullable=False)
    bundle_id = Column(Text, nullable=False)
    bundle_on_device = Column(Boolean, nullable=False)
    strategy = Column(Text, nullable=False)
    analyzers = Column(PickleType, nullable=False)

    device = relationship("Device", lazy="selectin")
    appium_server = relationship("AppiumServer", lazy="selectin")



import logging
import os

from ..exploration.exploration_session_options import ExplorationSessionOptions


class AnalyzerError(Exception):
    pass


class ExplorationAnalyzer:
    class Meta:
        platforms: list[str] = []

    exploration_session_options: ExplorationSessionOptions = None
    logger: logging.Logger = None

    def __init__(self, exploration_session_options: ExplorationSessionOptions):
        self.exploration_session_options = exploration_session_options
        self.logger = logging.getLogger(os.path.basename(exploration_session_options.log_directory))



from contextlib import asynccontextmanager
import logging
import os
import threading

from fastapi import Depends, FastAPI, Request
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
import uvicorn

from . import crud, models
from .common import grievios_base_path
from .database import SessionLocal, engine
from .exploration import exploration_worker
from .web import routers
from .web.dependencies import get_db, templates


@asynccontextmanager
async def lifespan(app: FastAPI):
    app.exploration_workers = []
    appium_servers = crud.get_appium_servers(db=SessionLocal())
    app.exploration_workers = [
        exploration_worker.ExplorationWorker(app, appium_server_id=server.id)
        for server in appium_servers
    ]
    for worker in app.exploration_workers:
        threading.Thread(target=worker.run).start()

    yield

    for worker in app.exploration_workers:
        worker.stop()


try:
    os.makedirs(grievios_base_path(), exist_ok=True)
    os.makedirs(grievios_base_path() / 'IPAs', exist_ok=True)
except OSError as e:
    logging.log(
        logging.ERROR,
        f"Failed to create GRIEviOS directories at ${grievios_base_path()}: {e}",
    )

models.Base.metadata.create_all(bind=engine)

print(os.path.dirname(__file__))

app = FastAPI(dependencies=[Depends(get_db)], lifespan=lifespan)
#app.mount("/static", StaticFiles(directory=os.path.relpath("grievios/web/static")), name="static")
app.mount("/static", StaticFiles(directory=os.path.join(os.path.abspath(os.path.dirname(__file__)), os.path.join("web", "static"))), name="static")
app.include_router(routers.devices.router)
app.include_router(routers.appium_servers.router)
app.include_router(routers.explorations.router)


@app.get("/", response_class=HTMLResponse)
async def root(request: Request):
    from .utils import cmdline_utils

    def check_util_availability(util: cmdline_utils.cmdline_util):
        path = None
        try:
            path = util.path()
        except cmdline_utils.cmdline_util.CmdlineUtilUnavailable:
            pass
        return path

    utils = {
        "ideviceinstaller": {"class": cmdline_utils.Ideviceinstaller, "required": True},
        "ideviceinfo": {"class": cmdline_utils.Ideviceinfo, "required": False},
        "idevice_id": {"class": cmdline_utils.IdeviceId, "required": False},
    }

    utils_info = {
        util: {
            "name": util,
            "required": util_info["required"],
            "path": check_util_availability(util_info["class"]),
        }
        for util, util_info in utils.items()
    }

    return templates.TemplateResponse(
        "index.html",
        {
            "request": request,
            "utilsInfo": utils_info,
            "grieviosDir": grievios_base_path(),
        },
    )


def start():
    uvicorn.run("grievios.main:app")



from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.encoders import jsonable_encoder
from fastapi.responses import HTMLResponse
from sqlalchemy.orm import Session

from ..dependencies import get_db, templates
from ... import crud, schemas
from ...utils.cmdline_utils import IdeviceId, Ideviceinfo
from ...utils.cmdline_utils.cmdline_util import CmdlineUtilUnavailableException

router = APIRouter()


@router.get("/devices", response_class=HTMLResponse)
async def devices(request: Request, db: Session = Depends(get_db)):
    devices = get_devices(db=db)
    return templates.TemplateResponse("devices.html", {"request": request, "devices": jsonable_encoder(devices)})


@router.get("/api/devices", response_model=list[schemas.Device])
def get_devices(skip: int = 0, limit: int = 100, db: Session = Depends(get_db)):
    devices = crud.get_devices(db, skip=skip, limit=limit)
    return devices


@router.get("/api/devices/connected", response_model=list[schemas.DeviceConnected])
def get_connected_devices():
    def get_ideviceinfos(device: schemas.DeviceConnected) -> schemas.DeviceConnected:
        infos = Ideviceinfo.run(device, ["DeviceName", "ProductName", "ProductVersion"])
        device.name = infos["DeviceName"]
        device.os = infos["ProductName"]
        device.os_version = infos["ProductVersion"]
        return device

    connected_devices: list[schemas.DeviceConnected] = []
    try:
        connected_devices = IdeviceId.run()
        connected_devices = [get_ideviceinfos(d) for d in connected_devices]
    except CmdlineUtilUnavailableException:
        pass
    return connected_devices


@router.get("/api/devices/{device_udid}", response_model=schemas.Device)
def get_device(device_udid: str, db: Session = Depends(get_db)):
    device = crud.get_device(db, device_udid=device_udid)
    if device is None:
        raise HTTPException(status_code=404, detail="Device not found")
    return device


@router.post("/api/devices", response_model=schemas.Device)
def add_device(device: schemas.DeviceCreate, db: Session = Depends(get_db)):
    db_device = crud.add_device(db, device=device)
    return db_device


@router.delete("/api/devices/{device_udid}")
def delete_device(device_udid: str, db: Session = Depends(get_db)):
    deleted_device = crud.delete_device(db, device_udid=device_udid)
    if deleted_device is None:
        raise HTTPException(status_code=404, detail=f"Device {device_udid} not found")



import logging
import os
import time

from appium import webdriver


class ExplorationStrategy:
    class Meta:
        platforms: list[str] = []

    wd: webdriver = None
    log_directory: os.path = None
    timeout: int = None
    logger: logging.Logger = None

    def __init__(self, wd: webdriver, log_directory: os.path, timeout: int = None):
        self.wd = wd
        self.log_directory = log_directory
        self.timeout = timeout
        self.logger = logging.getLogger(os.path.basename(log_directory))

    def run(self) -> None:
        stoptime: float = time.time() + self.timeout
        while time.time() < stoptime:
            self.interact()

    def interact(self):
        raise NotImplementedError



import subprocess


class CmdlineUtilUnavailableException(Exception):
    pass


class CmdlineUtil:
    name: str = ""

    @classmethod
    def path(cls) -> str:
        try:
            p: subprocess.CompletedProcess = subprocess.run(
                args=["which", cls.name],
                capture_output=True,
                check=True
            )
            path = p.stdout.rstrip().decode("UTF-8")
            if not path:
                raise CmdlineUtilUnavailableException
            return path
        except subprocess.CalledProcessError as e:
            raise CmdlineUtilUnavailableException(e)
        except FileNotFoundError:
            pass



import subprocess

from .cmdline_util import CmdlineUtil, CmdlineUtilUnavailableException


class Mitmdump(CmdlineUtil):
    name: str = "mitmdump"

    @staticmethod
    def subprocess(arguments: list[str] = None) -> subprocess:
        p: subprocess = None

        if arguments is None:
            arguments = []
        try:
            plog = b""
            p = subprocess.Popen(
                args=[Mitmdump.name] + arguments,
                stdout=subprocess.PIPE
            )
            while p.poll() is None and b"listening at" not in plog:
                plog += p.stdout.read(1)
        except subprocess.CalledProcessError as e:
            raise e
        except FileNotFoundError:
            raise CmdlineUtilUnavailableException
        return p



import subprocess

from .cmdline_util import CmdlineUtil, CmdlineUtilUnavailableException


class Tcpdump(CmdlineUtil):
    name: str = "tcpdump"
    p: subprocess = None

    def run(self, interface: str = None, output_file: str = None, sudo: bool = False):
        try:
            self.p = subprocess.Popen(
                args=
                (['sudo'] if sudo else []) +
                [Tcpdump.name] +
                (['-i', interface] if interface else []) +
                (['-w', output_file] if output_file else [])
            )
        except subprocess.CalledProcessError as e:
            raise e
        except FileNotFoundError as e:
            raise CmdlineUtilUnavailableException(e)

    def terminate(self):
        self.p.terminate()



from sqlalchemy.orm import Session

from . import models, schemas


""" Devices """


def get_devices(db: Session, skip: int = 0, limit: int = 100):
    return db.query(models.Device).all()


def get_device(db: Session, device_udid: str):
    return db.query(models.Device).filter(models.Device.udid == device_udid).first()


def add_device(db: Session, device: schemas.Device):
    db_device = models.Device(udid=device.udid, name=device.name, os=device.os, os_version=device.os_version)
    db.add(db_device)
    db.commit()
    db.refresh(db_device)
    return db_device


def delete_device(db: Session, device_udid: str):
    db_device = get_device(db, device_udid=device_udid)
    if not db_device is None:
        db.delete(db_device)
        db.commit()
    return db_device


""" Appium Servers """


def get_appium_servers(db: Session, skip: int = 0, limit: int = 100):
    return db.query(models.AppiumServer).all()


def get_appium_server_by_id(db: Session, id: int):
    return db.query(models.AppiumServer).filter(models.AppiumServer.id == id).first()


def add_appium_server(db: Session, appium_server: schemas.AppiumServerCreate):
    db_appium_server = models.AppiumServer(url=appium_server.url, port=appium_server.port, wdaBundleId=appium_server.wdaBundleId)
    db.add(db_appium_server)
    db.commit()
    db.refresh(db_appium_server)
    return db_appium_server


def delete_appium_server(db: Session, appium_server_id: int):
    db_appium_server = get_appium_server_by_id(db, appium_server_id)
    if db_appium_server:
        db.delete(db_appium_server)
        db.commit()
    return db_appium_server


""" Explorations """


def get_explorations(db: Session, skip: int = 0, limit: int = 100):
    return db.query(models.Exploration).all()


def get_exploration_by_id(db: Session, id: int):
    return db.query(models.Exploration).filter(models.Exploration.id == id).first()


def add_exploration(db: Session, exploration: schemas.ExplorationCreate):
    db_exploration = models.Exploration(**exploration.dict())
    db.add(db_exploration)
    db.commit()
    db.refresh(db_exploration)
    return db_exploration


def add_explorations(db: Session, explorations: list[schemas.Exploration]):
    db_explorations: list[schemas.Exploration] = []
    for exploration in explorations:
        db_exploration = models.Exploration(**exploration.dict(), device_udid=exploration.device.udid,
                                            appium_server_id=exploration.appium_server.id)
        db.add(db_exploration)
        db.commit()
        db.refresh(db_exploration)
        db_explorations.append(db_exploration)
    return db_explorations


def delete_exploration(db: Session, exploration_id: int):
    db_exploration = get_exploration_by_id(db, id=exploration_id)
    if db_exploration:
        db.delete(db_exploration)
        db.commit()
    return db_exploration



import random
from selenium.common.exceptions import NoSuchElementException

from .ios_strategy import iOSStrategy


class RandomButtonStrategy(iOSStrategy):

    def interact(self):
        graph = self.parse_json_source()
        self.logger.debug(f'graph: {graph}')
        buttons = list(filter(lambda el: el.__dict__.get('isVisible', '0') == '1' and el.__dict__.get('type', '') in self.XCUIElement.INTERACTIONS['tap'], graph.nodes))
        if not buttons:
            self.logger.info(f'No buttons found')
            return
        selected_button = random.choice(buttons)
        self.logger.debug(f'selected button: {selected_button}')
        try:
            appium_element = selected_button.get_wd_element(self.wd)
            appium_element.click()
        except NoSuchElementException:
            self.logger.warning(f'WebDriver couldnt find selected buttton')




import datetime
import os
import subprocess


from .ios_analyzer import iOSAnalyzer
from ...exploration.exploration_session_options import ExplorationSessionOptions
from ...utils.cmdline_utils.mitmdump import Mitmdump


class MitmDumpAnalyzer(iOSAnalyzer):
    mitmdump_process: subprocess = None

    def __init__(self, exploration_session_options: ExplorationSessionOptions):
        super().__init__(exploration_session_options)
        self.logger.info("Starting mitmdump process")
        self.mitmdump_process = Mitmdump.subprocess(
            arguments=
            ["-w",
             os.path.join(self.exploration_session_options.log_directory, datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S") + '.mitmdump')
             ]
        )
        self.logger.info("mitmdump started")
        self.logger.info("MitmDumpAnalyzer inited")

    def __del__(self):
        self.logger.info("Terminating mitmdump process")
        self.mitmdump_process.terminate()
        self.logger.info("MitmDumpAnalyzer deleted")

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.info("Terminating mitmdump process")
        self.mitmdump_process.terminate()
        self.logger.info("MitmDumpAnalyzer exited")



import datetime
import os

from .ios_analyzer import iOSAnalyzer
from ..exploration_analyzer import AnalyzerError
from ...exploration.exploration_session_options import ExplorationSessionOptions
from ...utils.cmdline_utils.cmdline_util import CmdlineUtilUnavailableException
from ...utils.cmdline_utils import rvictl
from ...utils.cmdline_utils.tcpdump import Tcpdump


class TcpdumpAnalyzer(iOSAnalyzer):
    p: Tcpdump = None

    def __init__(self, exploration_session_options: ExplorationSessionOptions):
        super().__init__(exploration_session_options)

        rvi_interface: str = None

        self.logger.info(f"Creating RVI for {self.exploration_session_options.device_udid}")
        try:
            self.rvi_interface = rvictl.Rvictl.start_device(udid=exploration_session_options.device_udid)
        except rvictl.RvictlError as e:
            self.logger.error(f"Failed to create RVI for {exploration_session_options.device_udid}")
            raise AnalyzerError
        except CmdlineUtilUnavailableException as e:
            self.logger.error(f"rvictl not avilable")
            raise AnalyzerError(e)
        self.logger.info(f"Created RVI: {self.rvi_interface}")
        self.logger.info("Starting Tcpdump")
        try:
            self.p = Tcpdump()
            self.p.run(interface=self.rvi_interface,
                       output_file=str(os.path.join(self.exploration_session_options.log_directory, datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S") + '.pcap')), # TODO: datetime.datetime.now().isoformat()
                       sudo=True)
        except Exception as e:
            self.logger.error(f"Failed to run tcpdump: {e}")
            raise e

    def __del__(self):
        try:
            self.logger.info("Terminating tcpdump")
            self.p.terminate()
        except Exception as e:
            self.logger.error(f"Failed to terminate tcpdump: {e}")
        try:
            self.logger.info(f"Removing RVI for {self.exploration_session_options.device_udid}")
            rvictl.Rvictl.stop_device(self.exploration_session_options.device_udid)
        except rvictl.RvictlError as e:
            self.logger.error(f"Failed to remove RVI for {self.exploration_session_options.device_udid}: {e}")
